// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file SparseVector.h */

#ifndef AMDIS_SPARSEVECTOR_H
#define AMDIS_SPARSEVECTOR_H

#include "MemoryManager.h"
#include "DOFVector.h"
#include <set>
#include <string>
#include <vector>

namespace AMDiS
{

  template<typename T> class SparseVector;

  template<typename T>
  void axpy(double a, SparseVector<T>& x, SparseVector<T>& y,
	    std::set<DegreeOfFreedom> &dofSet);

  template<typename T>
  void mv(DOFMatrix &a, 
	  SparseVector<T> &x,
	  SparseVector<T> &result,
	  std::set<DegreeOfFreedom> &dofSet,
	  bool add = false); 

  class FiniteElemSpace;

  template<typename T>
  class SparseVector : public DOFVectorBase<T>
  {
  public:
    MEMORY_MANAGED(SparseVector);

    SparseVector(const FiniteElemSpace *feSpace, const char *name)
      : DOFVectorBase<T>(feSpace, name),
	master_(NULL)
    {
      logToPhys_ = new std::map<DegreeOfFreedom, int>;
      physToLog_ = new std::vector<DegreeOfFreedom>;
    };

    SparseVector(const char *name, SparseVector *master)
      : DOFVectorBase<T>(master->getFESpace(), name),
	logToPhys_(master->logToPhys_),
	physToLog_(master->physToLog_),
	master_(master)
    {
      master_->clients_.push_back(this);
      vec_.resize(physToLog_->size(), 0.0);
    };

    virtual ~SparseVector() {
      if(master_) {
	typename std::vector<SparseVector<T>*>::iterator clientIt, clientEnd = master_->clients_.end();
	for(clientIt = master_->clients_.begin(); clientIt != clientEnd; ++clientIt) {
	  if(*clientIt == this) {
	    master_->clients_.erase(clientIt);
	    break;
	  }
	}
      } else {
	TEST_EXIT_DBG(clients_.empty())("delete clients first\n");
	delete logToPhys_;
	delete physToLog_;
      }
    };

    //void createSparseIndices(std::set<DegreeOfFreedom> &dofs);

    void copyToDOFVector(DOFVector<T> &dofVec);

    void copyToDOFVector(DOFVector<T> &dofVec,
			 std::set<DegreeOfFreedom> &dofs);

    void copyFromDOFVector(DOFVector<T> &dofVec,
			   std::set<DegreeOfFreedom> &dofs);

    void clear(bool eraseEntries = true);

    T norm();

    void copy(SparseVector &x, const std::set<DegreeOfFreedom> &dofSet);

    void copy(SparseVector &x, const std::vector<DegreeOfFreedom> &dofSet);

    void print();

    inline void synchronize() {
      vec_.resize(physToLog_->size(), 0.0);
    };

    virtual int getSize() const { 
      return static_cast<int>(physToLog_->size()); 
    };

    virtual void resize(int size) {};

    virtual void compressDOFIndexed(int first, int last, 
				    std::vector<DegreeOfFreedom> &newDOF) 
    {
      ERROR_EXIT("compress called\n");
    };

    virtual void freeDOFContent(int) {};

    virtual void refineInterpol(RCNeighbourList&, int);

    virtual void coarseRestrict(RCNeighbourList&, int);

    virtual typename std::vector<T>::iterator begin() {
      static typename std::vector<T>::iterator dummy;
      return dummy;
    };

    virtual typename std::vector<T>::iterator end() {
      static typename std::vector<T>::iterator dummy;
      return dummy;
    };

    virtual T& operator[](DegreeOfFreedom i);

    virtual const T& operator[](DegreeOfFreedom i) const;

    inline bool isSparseIndex(int i) {
      return  (*logToPhys_)[i] != 0;
    };

  protected:
    std::map<DegreeOfFreedom, int> *logToPhys_;

    std::vector<DegreeOfFreedom> *physToLog_;

    std::vector<T> vec_;

    SparseVector *master_;

    std::vector<SparseVector*> clients_;

    friend void axpy<>(double a, SparseVector<T>& x, SparseVector<T>& y,
		       std::set<DegreeOfFreedom> &dofSet);

    friend void mv<>(DOFMatrix &a,	
		     SparseVector<T> &x, 
		     SparseVector<T> &result,
		     std::set<DegreeOfFreedom> &dofSet,
		     bool add); 

  };

}

#include "SparseVector.hh"
#endif
